import logging
import time
from typing import TypedDict
from threading import Lock, Event

from .alg import (
    SETPOINT_UPDATE_T,
    UPDATE_T,
    calculate_jerk,
    get_initial_setpoint,
    has_reached_setpoint,
    move_to_setpoint,
    sanitize_fan_values,
    update_setpoint,
)
from .utils import (
    find_edge_temp,
    find_fans,
    find_tctl_temp,
    read_fan_speed,
    read_temp,
    write_fan_speed,
)

logger = logging.getLogger(__name__)


class FanInfo(TypedDict):
    tctl: str | None
    edge: str
    fans: list[str]


class FanData(TypedDict):
    a: float
    v: float
    t_target: int


class FanState(TypedDict):
    v_curr: float
    v_target: float
    v_target_pwm: int
    t_target: int
    v_rpm: list[int]
    t_junction: float | None
    t_edge: float
    fan_data: FanData
    in_setpoint: bool


def get_fan_info() -> FanInfo | None:
    tctl = find_tctl_temp()
    if tctl is None:
        logger.warning("Could not find tctl junction temperature.")

    edge = find_edge_temp()
    if edge is None:
        logger.error("Could not find edge temperature.")
        return None

    fans = find_fans()
    if not fans:
        logger.error("Could not find PWM controllable fans.")
        return None

    return {"tctl": tctl, "edge": edge, "fans": fans}


def calculate_fan_speed(
    temp: float, data: FanData | None, fan_curve: dict[int, float], junction: bool
) -> tuple[float, bool, FanData]:
    if data is None:
        # Initialize with best guess
        t_target = get_initial_setpoint(temp, fan_curve)
        v_curr = fan_curve[t_target]
        a_curr = 0
        return v_curr, False, {"a": a_curr, "v": v_curr, "t_target": t_target}

    # Get values and new temp target setpoint
    v_curr = data["v"]
    a_curr = data["a"]
    t_target = data["t_target"]
    t_target = update_setpoint(temp, t_target, fan_curve)

    # Pin values if we are in the setpoint
    if has_reached_setpoint(v_curr, a_curr, fan_curve[t_target]):
        a_curr = 0
        v_curr = fan_curve[t_target]
        return v_curr, True, {"a": a_curr, "v": v_curr, "t_target": t_target}

    v_target = fan_curve[t_target]
    jerk_accel, jerk_decel = calculate_jerk(t_target, v_target > v_curr, junction)
    v_new, a_new = move_to_setpoint(v_curr, a_curr, jerk_accel, jerk_decel, v_target)
    v_new, a_new = sanitize_fan_values(v_new, a_new)

    return v_new, False, {"a": a_new, "v": v_new, "t_target": t_target}


def set_fans_to_pwm(enable: bool, fan_info: FanInfo):
    for _, fn_enable, _, legacy in fan_info["fans"]:
        with open(fn_enable, "w") as f:
            if enable:
                f.write("1")
            else:
                f.write("0" if legacy else "2")


def update_fan_speed(
    state: FanState | None,
    fan_info: FanInfo,
    fan_curve: dict[int, float],
    junction: bool,
    observe_only: bool = False,
) -> tuple[bool, FanState]:
    t_edge = read_temp(fan_info["edge"])
    if fan_info["tctl"] is not None:
        t_junction = read_temp(fan_info["tctl"])
    else:
        t_junction = None

    t_curr = t_junction if junction else t_edge
    assert t_curr is not None, "Current temperature cannot be None"
    
    data = state["fan_data"] if state else None
    v_curr, in_setpoint, data = calculate_fan_speed(t_curr, data, fan_curve, junction)

    v_curr_int = min(255, max(0, int(v_curr * 255)))
    if not observe_only:
        if state is None or state["v_target_pwm"] != v_curr_int:
            for v_fn, _, _, _ in fan_info["fans"]:
                write_fan_speed(v_fn, v_curr_int)

    fan_speeds = [read_fan_speed(rpm_fn) for _, _, rpm_fn, _ in fan_info["fans"] if rpm_fn]
    return (
        in_setpoint,
        {
            "v_curr": v_curr,
            "v_target": fan_curve[data["t_target"]],
            "v_target_pwm": v_curr_int,
            "v_rpm": fan_speeds,
            "t_junction": t_junction,
            "t_edge": t_edge,
            "t_target": data["t_target"],
            "fan_data": data,
            "in_setpoint": in_setpoint,
        },
    )


def fan_worker(
    info: FanInfo,
    should_exit: Event,
    lock: Lock,
    fan_curve: dict[int, float],
    state: FanState,
    junction: Event,
):
    try:
        set_fans_to_pwm(True, info)
        while not should_exit.is_set():
            with lock:
                state_tmp = state or None  # First time state will be empty
                in_setpoint, state_tmp = update_fan_speed(
                    state_tmp, info, fan_curve, junction.is_set()
                )
                state.update(state_tmp)
            time.sleep(SETPOINT_UPDATE_T if in_setpoint else UPDATE_T)
    except Exception as e:
        logger.error(f"Fan worker failed:\n{e}")
    finally:
        set_fans_to_pwm(False, info)


def fan_pwm_tester(normal_curve: bool = True, observe_only: bool = False):
    fan_info = get_fan_info()
    if fan_info is None:
        return

    if normal_curve:
        fan_curve = {
            50: 0.3,
            60: 0.35,
            70: 0.4,
            80: 0.5,
            85: 0.6,
            90: 0.8,
            100: 0.9,
        }
        fan_curve = {
            40: 0.2,
            45: 0.3,
            50: 0.4,
            55: 0.45,
            60: 0.55,
            65: 0.7,
            70: 0.8,
            80: 0.85,
            90: 0.9,
            100: 1,
        }
    else:
        fan_curve = {
            50: 0.3,
            60: 0.3,
            65: 0.9,
            80: 0.9,
            90: 0.9,
            100: 1,
        }

    try:
        if not observe_only:
            set_fans_to_pwm(True, fan_info)

        MAX_FAN = 5300

        state = None
        for i in range(10000000):
            in_setpoint, state = update_fan_speed(
                state, fan_info, fan_curve, False, observe_only=observe_only
            )

            print(
                f"\n> {i:05d}: {'in setpoint' if in_setpoint else 'updating'}{' (observe)' if observe_only else ''}"
            )
            print(
                f"  Junction: {state['t_junction']:.2f}C, Edge: {state['t_edge']:.2f}C"
            )
            print(
                f"  Current: {state['v_curr']*100:.1f}%, Target: {state['v_target']*100:.1f}%"
            )
            print(
                f"  Fan speeds: {' '.join(map(lambda rpm: f"{rpm:4d}rpm/{MAX_FAN}rpm ({100*rpm/MAX_FAN:.1f}%)", state['v_rpm']))}"
            )
            time.sleep(SETPOINT_UPDATE_T if in_setpoint else UPDATE_T)
    except KeyboardInterrupt:
        print("Exiting fan test.")
    finally:
        set_fans_to_pwm(False, fan_info)



#!/usr/bin/env python3

import fcntl
import os
import sys

import dbus
import dbus.mainloop.glib
import dbus.service
from gi.repository import GLib

BASE_NAME = "org.freedesktop.UPower.PowerProfiles"
BASE_PATH = "/org/freedesktop/UPower/PowerProfiles"
LEGACY_NAME = "net.hadess.PowerProfiles"
LEGACY_PATH = "/net/hadess/PowerProfiles"
XML_PATH = "power-profiles-daemon.dbus.xml.in"

SUPPORTED_PROFILES = {
    "power-saver": "power",
    "balanced": "balanced",
    "performance": "performance",
}
SUPPORTED_PROFILES_REVERSE = {v: k for k, v in SUPPORTED_PROFILES.items()}


def load_introspect(legacy=False):
    """Returns the yaml data of a file in the relative dir provided."""
    import inspect
    import os

    script_fn = inspect.currentframe().f_back.f_globals["__file__"]  # type: ignore
    dirname = os.path.dirname(script_fn)
    with open(os.path.join(dirname, XML_PATH), "r") as f:
        base = f.read()
        return base.replace(
            "@dbus_iface@", LEGACY_NAME if legacy else BASE_NAME
        ).replace("@dbus_path@", LEGACY_PATH if legacy else BASE_PATH)


def iface(legacy: bool):
    return LEGACY_NAME if legacy else BASE_NAME


def gpath(legacy: bool):
    return LEGACY_PATH if legacy else BASE_PATH


def create_interface(legacy: bool):
    class HhdPpd(dbus.service.Object):

        def __init__(self, conn=None):
            self.profile_holds = []
            self.actions = []
            self.profile = "power-saver"  # next(iter(SUPPORTED_PROFILES))

            super().__init__(conn, gpath(legacy), None)

        @dbus.service.method(
            dbus.INTROSPECTABLE_IFACE, in_signature="", out_signature="s"
        )
        def Introspect(self):
            return load_introspect(legacy)

        @dbus.service.method(
            dbus.PROPERTIES_IFACE, in_signature="ss", out_signature="v"
        )
        def Get(self, interface_name, property_name):
            return self.GetAll(interface_name)[property_name]

        @dbus.service.method(
            dbus.PROPERTIES_IFACE,
            in_signature="s",
            out_signature="a{sv}",
            sender_keyword="sender",
        )
        def GetAll(self, interface_name, sender=None):
            if interface_name == iface(legacy):
                return {
                    "Actions": ["trickle_charge"],
                    "ActiveProfile": self.profile,
                    "ActiveProfileHolds": dbus.Array(self.profile_holds, signature="u"),
                    "PerformanceDegraded": "",
                    "PerformanceInhibited": "",
                    "Profiles": [
                        {
                            "Profile": dbus.String(p, variant_level=1),
                            "CpuDriver": dbus.String("amd_pstate", variant_level=1),
                            "PlatformDriver": dbus.String(
                                "platform_profile", variant_level=1
                            ),
                            "Driver": dbus.String("multiple", variant_level=1),
                        }
                        for p in SUPPORTED_PROFILES
                    ],
                    "Version": "0.21",
                }
            else:
                raise dbus.exceptions.DBusException(
                    "com.example.UnknownInterface",
                    "Handheld daemon does not implement the %s interface."
                    % interface_name,
                )

        @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature="ssv")
        def Set(self, interface_name, property_name, new_value):
            # validate the property name and value, update internal state…
            self.PropertiesChanged(interface_name, {property_name: new_value}, [])

        @dbus.service.signal(dbus.PROPERTIES_IFACE, signature="sa{sv}as")
        def PropertiesChanged(
            self, interface_name, changed_properties, invalidated_properties
        ):
            if interface_name != iface(legacy):
                return

            for k, v in changed_properties.items():
                if not k == "ActiveProfile":
                    continue
                if v not in SUPPORTED_PROFILES:
                    continue
                np = SUPPORTED_PROFILES[v]
                if np == self.profile:
                    continue
                print(np, flush=True)

        @dbus.service.method(iface(legacy), in_signature="sss", out_signature="u")
        def HoldProfile(self, profile: str, reason: str, application_id: str):
            # TODO
            return 1

        @dbus.service.method(iface(legacy), in_signature="u", out_signature="")
        def ReleaseProfile(self, handle: int):
            # TODO
            self.ProfileReleased(handle)

        @dbus.service.signal(iface(legacy), signature="u")
        def ProfileReleased(self, handle: int):
            # TODO
            return handle

        def set_profile(self, profile):
            if profile not in SUPPORTED_PROFILES_REVERSE:
                return

            self.profile = SUPPORTED_PROFILES_REVERSE[profile]
            self.PropertiesChanged(iface(legacy), {"ActiveProfile": self.profile}, [])

    return HhdPpd

def update_profile(services):
    for line in sys.stdin:
        if not line:
            break
        profile = line.strip()
        
        for s in services:
            s.set_profile(profile)
    return True

if __name__ == "__main__":
    mainloop = None
    try:
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

        # set sys.stdin non-blocking
        orig_fl = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin, fcntl.F_SETFL, orig_fl | os.O_NONBLOCK)

        legacy = False
        session_bus = dbus.SystemBus()
        
        claimed = False
        try:
            name1 = dbus.service.BusName(iface(False), session_bus)
            claimed = True
        except Exception:
            pass
        
        try:
            name2 = dbus.service.BusName(iface(True), session_bus)
            claimed = True
        except Exception:
            pass
        
        if not claimed:
            sys.exit(1)

        services = [
            create_interface(False)(session_bus),
            create_interface(True)(session_bus),
        ]

        GLib.timeout_add(100, lambda: update_profile(services))

        mainloop = GLib.MainLoop()
        mainloop.run()
    except KeyboardInterrupt:
        if mainloop:
            mainloop.quit()



from typing import TypedDict

from .alib import A, D, DeviceParams, AlibParams


class DevicePreset(TypedDict):
    tdp_limit: int
    slow_limit: int
    fast_limit: int
    slow_time: int
    stapm_time: int
    temp_target: int
    fan_curve: dict[int, float] | None


class DevideProfile(TypedDict):
    quiet: DevicePreset
    balanced: DevicePreset
    performance: DevicePreset
    # Turbo is custom with max tdp values
    turbo: DevicePreset

    platform_profile_map: dict[str, int]
    ppd_balanced_min: int
    ppd_performance_min: int

    alib: dict[str, AlibParams]
    dev: dict[str, DeviceParams]


class DeviceTDP(TypedDict):
    quiet: int
    balanced_min: int
    balanced: int
    performance_min: int
    performance: int
    performance_dc: int | None
    min_tdp: int
    max_tdp_dc: int | None
    max_tdp: int
    max_tdp_boost: int | None
    max_tdp_oc: int
    supports_cycle: bool | None


class DeviceTDPv2(TypedDict):
    quiet: tuple[int, int, int | None]
    balanced_min: int
    balanced: tuple[int, int, int | None]
    performance_min: int
    performance: tuple[int, int, int | None]
    performance_dc: int | None
    min_tdp: int
    max_tdp_dc: int | None
    max_tdp: int
    max_tdp_oc: int
    max_tdp_sppt: int | None
    max_tdp_fppt: int | None
    supports_cycle: bool | None


# internal name for ppd, platform_profile choices, min TDP for
# profile to apply, tdp target to apply when selecting profile
ENERGY_MAP = [
    ("power", ["low-power", "quiet"], 0, 8),
    ("balanced", ["balanced"], 13, 15),
    ("performance", ["performance"], 20, 25),
]
ENERGY_MAP_18W = [
    ("low-power", ["low-power", "quiet"], 0, 5),
    ("balanced", ["balanced"], 8, 12),
    ("performance", ["performance"], 13, 18),
]
ENERGY_MAP_WIN5 = [
    ("power", ["low-power", "quiet"], 0, 15),
    ("balanced", ["balanced"], 30, 25),
    ("performance", ["performance"], 50, 60),
]
ENERGY_MAP_AIMAX = [
    ("power", ["low-power", "quiet"], 0, 15),
    ("balanced", ["balanced"], 30, 50),
    ("performance", ["performance"], 50, 90),
]

ALIB_PARAMS = {
    # TDPs
    "stapm_limit": A(0x05, 0, 54, 1000),
    "fast_limit": A(0x06, 0, 54, 1000),
    "slow_limit": A(0x07, 0, 54, 1000),
    "skin_limit": A(0x2E, 0, 54, 1000),
    # Times
    "slow_time": A(0x08, 0, 30),
    "stapm_time": A(0x01, 0, 300),
    # Temp
    "temp_target": A(0x03, 0, 105),
}

ALIB_PARAMS_AIMAX = {
    # TDPs
    "stapm_limit": A(0x05, 0, 120, 1000),
    "fast_limit": A(0x06, 0, 120, 1000),
    "slow_limit": A(0x07, 0, 120, 1000),
    "skin_limit": A(0x2E, 0, 120, 1000),
    # Times
    "slow_time": A(0x08, 0, 30),
    "stapm_time": A(0x01, 0, 300),
    # Temp
    "temp_target": A(0x03, 0, 105),
}

ALIB_PARAMS_5040: dict[str, AlibParams] = ALIB_PARAMS
ALIB_PARAMS_7040: dict[str, AlibParams] = ALIB_PARAMS
ALIB_PARAMS_6040: dict[str, AlibParams] = ALIB_PARAMS
ALIB_PARAMS_8040: dict[str, AlibParams] = ALIB_PARAMS
ALIB_PARAMS_HX370: dict[str, AlibParams] = ALIB_PARAMS

DEV_PARAMS_30W: dict[str, DeviceParams] = {
    "stapm_limit": D(0, 4, 15, 30, 40),
    "skin_limit": D(0, 4, 15, 30, 40),
    "slow_limit": D(0, 4, 20, 32, 43),
    "fast_limit": D(0, 4, 25, 41, 50),
    # Times
    "slow_time": D(5, 5, 10, 10, 10),
    "stapm_time": D(100, 100, 100, 200, 200),
    # Temp
    "temp_target": D(60, 70, 85, 90, 100),
}

DEV_PARAMS_18W: dict[str, DeviceParams] = {
    "stapm_limit": D(0, 5, 15, 18, 22),
    "skin_limit": D(0, 5, 15, 18, 22),
    "slow_limit": D(0, 5, 15, 18, 22),
    "fast_limit": D(0, 5, 15, 20, 25),
    # Times
    "slow_time": D(5, 5, 10, 10, 10),
    "stapm_time": D(100, 100, 100, 200, 200),
    # Temp
    "temp_target": D(60, 70, 85, 90, 100),
}
DEV_PARAMS_25W: dict[str, DeviceParams] = {
    "stapm_limit": D(0, 4, 15, 25, 32),
    "skin_limit": D(0, 4, 15, 25, 32),
    "slow_limit": D(0, 4, 20, 27, 35),
    "fast_limit": D(0, 4, 25, 30, 37),
    # Times
    "slow_time": D(5, 5, 10, 10, 10),
    "stapm_time": D(100, 100, 100, 200, 200),
    # Temp
    "temp_target": D(60, 70, 85, 90, 100),
}
DEV_PARAMS_28W: dict[str, DeviceParams] = {
    "stapm_limit": D(0, 4, 15, 28, 32),
    "skin_limit": D(0, 4, 15, 28, 32),
    "slow_limit": D(0, 4, 20, 30, 35),
    "fast_limit": D(0, 4, 25, 32, 37),
    # Times
    "slow_time": D(5, 5, 10, 10, 10),
    "stapm_time": D(100, 100, 100, 200, 200),
    # Temp
    "temp_target": D(60, 70, 85, 90, 100),
}
DEV_PARAMS_AIMAX: dict[str, DeviceParams] = {
    "stapm_limit": D(0, 4, 25, 120, 120),
    "skin_limit": D(0, 4, 25, 120, 120),
    "slow_limit": D(0, 4, 27, 120, 120),
    "fast_limit": D(0, 4, 40, 120, 120),
    # Times
    "slow_time": D(5, 5, 10, 10, 10),
    "stapm_time": D(100, 100, 100, 200, 200),
    # Temp
    "temp_target": D(60, 70, 85, 90, 100),
}

DEV_PARAMS_WIN5: dict[str, DeviceParams] = {
    "stapm_limit": D(0, 4, 25, 85, 100),
    "skin_limit": D(0, 4, 25, 85, 100),
    "slow_limit": D(0, 4, 27, 85, 100),
    "fast_limit": D(0, 4, 40, 85, 100),
    # Times
    "slow_time": D(5, 5, 10, 10, 10),
    "stapm_time": D(100, 100, 100, 200, 200),
    # Temp
    "temp_target": D(60, 70, 85, 90, 100),
}


DEV_PARAMS_5000: dict[str, DeviceParams] = DEV_PARAMS_25W
DEV_PARAMS_6000: dict[str, DeviceParams] = DEV_PARAMS_30W
DEV_PARAMS_7040: dict[str, DeviceParams] = DEV_PARAMS_30W
DEV_PARAMS_8040: dict[str, DeviceParams] = DEV_PARAMS_30W
DEV_PARAMS_HX370: dict[str, DeviceParams] = DEV_PARAMS_30W
DEV_PARAMS_LEGO = DEV_PARAMS_30W

DEV_DATA: dict[
    str,
    tuple[
        dict[str, DeviceParams],
        dict[str, AlibParams],
        bool,
        list[tuple[str, list[str], int, int]],
    ],
] = {
    "NEO-01": (DEV_PARAMS_28W, ALIB_PARAMS_7040, False, ENERGY_MAP),
    "V3": (DEV_PARAMS_28W, ALIB_PARAMS_8040, False, ENERGY_MAP),
    "83E1": (DEV_PARAMS_LEGO, ALIB_PARAMS_7040, False, ENERGY_MAP),
    "ONEXPLAYER F1Pro": (DEV_PARAMS_HX370, ALIB_PARAMS_HX370, False, ENERGY_MAP),
    "ONEXPLAYER F1 EVA-02": (DEV_PARAMS_HX370, ALIB_PARAMS_HX370, False, ENERGY_MAP),
    # GPD Devices are 28W max
    "G1618-04": (DEV_PARAMS_28W, ALIB_PARAMS_7040, False, ENERGY_MAP),
    "G1617-01": (DEV_PARAMS_28W, ALIB_PARAMS_7040, False, ENERGY_MAP),
    "G1619-04": (DEV_PARAMS_28W, ALIB_PARAMS_7040, False, ENERGY_MAP),
    "G1619-05": (DEV_PARAMS_28W, ALIB_PARAMS_7040, False, ENERGY_MAP),
    "G1618-05": (DEV_PARAMS_WIN5, ALIB_PARAMS_AIMAX, False, ENERGY_MAP_AIMAX),
}

CPU_DATA: dict[
    str,
    tuple[
        dict[str, DeviceParams],
        dict[str, AlibParams],
        list[tuple[str, list[str], int, int]],
    ],
] = {
    "AMD Ryzen Z1 Extreme": (DEV_PARAMS_7040, ALIB_PARAMS_7040, ENERGY_MAP),
    "AMD Ryzen Z1": (DEV_PARAMS_7040, ALIB_PARAMS_7040, ENERGY_MAP),
    # Ayaneo AIR Pro, max is 18W
    "AMD Ryzen 5 5560U": (DEV_PARAMS_18W, ALIB_PARAMS_5040, ENERGY_MAP_18W),
    # 28W works fine, 30W is pushing it
    "AMD Ryzen 7 5700U": (DEV_PARAMS_5000, ALIB_PARAMS_5040, ENERGY_MAP),
    "AMD Ryzen 7 5800U": (DEV_PARAMS_5000, ALIB_PARAMS_5040, ENERGY_MAP),
    # GPD Win 4
    # model name    : AMD Ryzen 7 6800U with Radeon Graphics
    "AMD Ryzen 7 6600U": (DEV_PARAMS_6000, ALIB_PARAMS_6040, ENERGY_MAP),
    "AMD Ryzen 7 6800U": (DEV_PARAMS_6000, ALIB_PARAMS_6040, ENERGY_MAP),
    "AMD Ryzen 7 7840U": (DEV_PARAMS_7040, ALIB_PARAMS_7040, ENERGY_MAP),
    "AMD Ryzen 7 8840U": (DEV_PARAMS_8040, ALIB_PARAMS_8040, ENERGY_MAP),
    # AMD Athlon Silver 3050e (Win600, will it support tdp?)
    "AMD Ryzen AI 9 HX 370": (DEV_PARAMS_HX370, ALIB_PARAMS_HX370, ENERGY_MAP),
    "AMD Ryzen AI HX 360": (DEV_PARAMS_HX370, ALIB_PARAMS_HX370, ENERGY_MAP),
    # AMD RYZEN AI MAX+ 395 w/ Radeon 8060S
    "AMD RYZEN AI MAX+ 395": (DEV_PARAMS_AIMAX, ALIB_PARAMS_AIMAX, ENERGY_MAP_AIMAX),
}

ALLY_DATA: DeviceTDP = {
    "quiet": 10,
    "balanced_min": 13,
    "balanced": 15,
    "performance_min": 20,
    "performance_dc": 25,
    "performance": 30,
    "min_tdp": 5,
    "max_tdp_dc": 25,
    "max_tdp": 30,
    "max_tdp_oc": 50,
    "max_tdp_boost": None,
    "supports_cycle": True,
}

ALLYX_DATA: DeviceTDP = {
    "quiet": 13,
    "balanced_min": 15,
    "balanced": 17,
    "performance_min": 22,
    "performance_dc": 25,
    "performance": 30,
    "min_tdp": 5,
    "max_tdp_dc": 25,
    "max_tdp": 30,
    "max_tdp_oc": 50,
    "max_tdp_boost": None,
    "supports_cycle": True,
}


XBOX_ALLY_DATA: DeviceTDP = {
    "quiet": 6,
    "balanced_min": 8,
    "balanced": 15,
    "performance_min": 17,
    "performance_dc": 20,
    "performance": 20,
    "min_tdp": 4,
    "max_tdp_dc": 20,
    "max_tdp": 20, # SPPT 20 FPPT 25
    "max_tdp_oc": 25,
    "max_tdp_boost": None,
    "supports_cycle": True,
}

XBOX_ALLYX_DATA: DeviceTDP = {
    "quiet": 13,
    "balanced_min": 15,
    "balanced": 17,
    "performance_min": 20,
    "performance_dc": 35,
    "performance": 35,
    "min_tdp": 4,
    "max_tdp_dc": 35,
    "max_tdp": 35,
    "max_tdp_oc": 50,
    "max_tdp_boost": None,
    "supports_cycle": True,
}

Z1_DATA: DeviceTDP = {
    "quiet": 40,
    "balanced_min": 42,
    "balanced": 45,
    "performance_min": 50,
    "performance_dc": 54,
    "performance": 65,
    "min_tdp": 5,
    "max_tdp_dc": 54,
    "max_tdp": 65,
    "max_tdp_oc": 90,
    "max_tdp_boost": None,
    "supports_cycle": False,
}

ASUS_DATA: dict[str, DeviceTDP] = {
    "ROG Ally RC71L": ALLY_DATA,
    "ROG Ally X RC72L": ALLYX_DATA,
    "ROG Xbox Ally RC73Y": XBOX_ALLY_DATA,
    "ROG Xbox Ally X RC73X": XBOX_ALLYX_DATA,
    "ROG Flow Z13 GZ302": Z1_DATA,
}

CLAW_DATA: DeviceTDPv2 = {
    "quiet": (20, 20, None),
    "balanced_min": 21,
    "balanced": (30, 30, None),
    "performance_min": 36,
    "performance_dc": 35,
    "performance": (35, 35, None),
    "min_tdp": 20,
    "max_tdp_dc": 35,
    "max_tdp": 43,
    "max_tdp_oc": 45,
    "max_tdp_sppt": 45,
    "max_tdp_fppt": None,
    "supports_cycle": True,
}

CLAW_AI_DATA: DeviceTDPv2 = {
    "quiet": (8, 37, None),
    "balanced_min": 9,
    "balanced": (12, 37, None),
    "performance_min": 16,
    "performance_dc": 30,
    "performance": (30, 37, None),
    "min_tdp": 8,
    "max_tdp_dc": 30,
    "max_tdp": 30,
    "max_tdp_oc": 30,
    "max_tdp_sppt": 37,
    "max_tdp_fppt": None,
    "supports_cycle": True,
}

CLAW_8A_DATA: DeviceTDPv2 = {
    "quiet": (15, 28, 33),
    "balanced_min": 17,
    "balanced": (20, 33, 43),
    "performance_min": 23,
    "performance_dc": 28,
    "performance": (28, 45, 55),
    "min_tdp": 8,
    "max_tdp_dc": 35,
    "max_tdp": 35,
    "max_tdp_oc": 35,
    "max_tdp_sppt": 37,
    "max_tdp_fppt": 55,
    "supports_cycle": True,
}

MSI_DATA: dict[str, DeviceTDPv2] = {
    "MS-1T41": CLAW_DATA,
    "MS-1T42": CLAW_AI_DATA,
    "MS-1T52": CLAW_AI_DATA,
    "MS-1T8K": CLAW_8A_DATA,
}



import os
import subprocess
from typing import Literal
import shutil

import time
import signal
from hhd.plugins import Context, HHDPlugin, load_relative_yaml
from hhd.plugins.conf import Config
from threading import Event
import logging

from hhd.plugins.plugin import Emitter

logger = logging.getLogger(__name__)

PROFILES = ["performance", "balanced", "power-saver"]
PTYPE = Literal["performance", "balanced", "power-saver"]


def set_power_profile(profile):
    try:
        busctl = shutil.which("busctl")
        if not busctl:
            return None
        subprocess.run(
            [
                busctl,
                "set-property",
                "org.freedesktop.UPower.PowerProfiles",
                "/org/freedesktop/UPower/PowerProfiles",
                "org.freedesktop.UPower.PowerProfiles",
                "ActiveProfile",
                "s",
                profile,
            ],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except Exception as e:
        logger.warning(f"Could set power profile to '{profile}': {e}")


def get_current_power_profile() -> PTYPE | None:
    try:
        busctl = shutil.which("busctl")
        if not busctl:
            return None
        res = subprocess.run(
            [
                busctl,
                "get-property",
                "org.freedesktop.UPower.PowerProfiles",
                "/org/freedesktop/UPower/PowerProfiles",
                "org.freedesktop.UPower.PowerProfiles",
                "ActiveProfile",
            ],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        profile = res.stdout.decode().strip()
        for p in PROFILES:
            if p in profile:
                return p  # type: ignore
    except Exception:
        pass
    logger.warning(f"Could not read power profile. Disabling profile support.")


class GeneralPowerPlugin(HHDPlugin):

    def __init__(
        self,
    ) -> None:
        self.name = f"adjustor_general"
        self.priority = 8
        self.log = "gpow"
        self.last_check = None
        self.target = None
        self.old_sched = None
        self.sched_proc = None
        self.ppd_supported = None
        self.ovr_enabled = False
        self.should_exit = Event()
        self.t_sys = None
        self.currentTarget = None

    def open(self, emit: Emitter, context: Context):
        self.emit = emit
        self.ppd_supported = get_current_power_profile() is not None

        # SchedExt
        self.sets = load_relative_yaml("./settings.yml")
        self.avail_scheds = {}
        self.avail_pretty = {}
        kernel_supports = os.path.isfile("/sys/kernel/sched_ext/state")
        if kernel_supports:
            for sched, pretty in self.sets["children"]["sched"]["options"].items():
                if sched == "disabled":
                    self.avail_pretty[sched] = pretty
                    continue

                exe = shutil.which(sched)
                if exe:
                    self.avail_scheds[sched] = exe
                    self.avail_pretty[sched] = pretty

        if not self.ppd_supported:
            del self.sets["children"]["profile"]

        if self.avail_scheds:
            self.sets["children"]["sched"]["options"] = self.avail_pretty
        else:
            del self.sets["children"]["sched"]

    def settings(self):
        sets = self.sets

        self.logged_boost = True
        return {
            "tdp": {"general": sets},
        }

    def update(self, conf: Config):
        # Handle ppd
        if self.ppd_supported:
            curr = time.time()
            new_profile = conf.get("tdp.general.profile", self.target)
            if new_profile != self.target and new_profile and self.target:
                logger.info(f"Setting power profile to '{new_profile}'")
                set_power_profile(new_profile)
                self.target = new_profile
            elif not self.last_check or curr - self.last_check > 2:
                # Update profile every 2 seconds
                self.last_check = curr
                self.target = get_current_power_profile()
                if self.target is None:
                    self.ppd_supported = False
                    logger.info(f"Power profile support seems to be gone. Disabling.")
                    self.emit({"type": "settings"})
                elif self.target != conf["tdp.general.profile"].to(str):
                    conf["tdp.general.profile"] = self.target

        # Handle sched
        if self.avail_scheds:
            # Check health and print error
            if self.sched_proc and self.sched_proc.poll():
                err = self.sched_proc.poll()
                self.sched_proc = None
                logger.error(
                    f"Scheduler from sched_ext '{self.old_sched}' closed with error code: {err}"
                )

            new_sched = conf.get("tdp.general.sched", "disabled")
            if new_sched != self.old_sched:
                self.close_sched()
                self.old_sched = new_sched
                if new_sched and new_sched != "disabled":
                    logger.info(f"Starting sched_ext scheduler '{new_sched}'")
                    self.sched_proc = subprocess.Popen(
                        self.avail_scheds[new_sched],
                        stderr=subprocess.DEVNULL,
                        stdout=subprocess.DEVNULL,
                    )

    def close_sched(self):
        if self.sched_proc is not None:
            logger.info(f"Closing sched_ext scheduler '{self.old_sched}'.")
            self.sched_proc.send_signal(signal.SIGINT)
            self.sched_proc.wait()
            self.sched_proc = None

    def close(self):
        self.close_sched()
        if self.t_sys:
            self.should_exit.set()
            self.t_sys.join()



from .acpi import call
import logging
from typing import NamedTuple, Literal


class AlibParams(NamedTuple):
    id: int
    min: int
    max: int
    scale: int = 1


class DeviceParams(NamedTuple):
    min: int | None
    smin: int | None
    default: int | None
    smax: int | None
    max: int | None


Limit = Literal["device", "expanded", "cpu", "unlocked"]
A = AlibParams
D = DeviceParams

logger = logging.getLogger(__name__)


def alib(
    params: dict[str, int],
    cpu: dict[str, A],
    limit: Limit = "device",
    dev: dict[str, D] = {},
):
    length = 2
    data = bytearray()
    info = f"Sending SMU command with {len(params)} parameters:"
    for name, val in params.items():
        length += 5
        if name not in cpu:
            logger.error(
                f"Command '{name}' not found in instructions:\n{cpu}\nSkipping ALIB command."
            )
            return False

        cmd, cmin, cmax, scale = cpu[name]
        if limit != "unlocked" and (val < cmin or val > cmax):
            logger.error(f"Value {val} violates APU limit for {name}: {cpu[name]}")
            return False

        if dev and name in dev:
            dmin, smin, _, smax, dmax = dev[name]
            if limit == "device" and (
                (dmin is not None and val < dmin) or (dmax is not None and val > dmax)
            ):
                logger.error(
                    f"Value {val} violates device limit for {name}: {dev[name]}"
                )
                return False
            if limit in ("device", "expanded") and (
                (smin is not None and val < smin) or (smax is not None and val > smax)
            ):
                logger.error(
                    f"Value {val} violates expanded device limit for {name}: {dev[name]}"
                )
                return False

        data.append(cmd)
        data.extend(
            int.to_bytes(scale * val, length=4, byteorder="little", signed=False)
        )
        info += f"\n - {name:>12s} (0x{cmd:02x}): {val}"

    b_length = int.to_bytes(length, length=2, byteorder="little", signed=False)
    logger.info(info)
    return call(r"\_SB.ALIB", [0x0C, b_length + data])



import logging

logger = logging.getLogger(__name__)


def get_platform_choices():
    try:
        with open("/sys/firmware/acpi/platform_profile_choices", "r") as f:
            return f.read().strip().split(" ")
    except Exception:
        logger.info(
            f"Could not enumerate platform profile choices. Disabling platform profile."
        )
        return None


def set_platform_profile(prof: str):
    try:
        logger.info(f"Setting platform profile to '{prof}'")
        with open("/sys/firmware/acpi/platform_profile", "w") as f:
            f.write(prof)
        return True
    except Exception as e:
        logger.error(f"Could not set platform profile with error:\n{e}")
        return False


def get_platform_profile():
    try:
        with open("/sys/firmware/acpi/platform_profile", "r") as f:
            return f.read().replace("\n", "")
    except Exception as e:
        logger.error(f"Could not read platform profile with error:\n{e}")
        return None



import logging
import os
import select
import time
from threading import Event as TEvent

from hhd.controller import DEBUG_MODE, Multiplexer
from hhd.controller.lib.hide import unhide_all
from hhd.controller.physical.evdev import XBOX_BUTTON_MAP
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev, enumerate_evs
from hhd.controller.physical.hidraw import GenericGamepadHidraw
from hhd.controller.physical.imu import CombinedImu, HrtimerTrigger
from hhd.controller.virtual.uinput import UInputDevice
from hhd.i18n import _
from hhd.plugins import Config, Context, Emitter, get_gyro_state, get_outputs

from .const import (
    AYA3_INIT,
    AYA_CHECK,
    AYA_CUSTOM,
    DEFAULT_MAPPINGS,
    AYA_SAVE,
    get_cfg_commands,
)


def to_bytes(s: str):
    return bytes.fromhex(s.replace(" ", ""))


FIND_DELAY = 0.1
ERROR_DELAY = 0.3
LONGER_ERROR_DELAY = 3
LONGER_ERROR_MARGIN = 1.3

logger = logging.getLogger(__name__)

GAMEPAD_VID = 0x045E
GAMEPAD_PID = 0x028E

KBD_VID = 0x0001
KBD_PID = 0x0001

AYA_VID = 0x1C4F
AYA_PID = 0x0002

AYA_TIMEOUT = 300
AYA_SAVE_TIMEOUT = 1500
AYA_ATTEMPTS = 3
AYA_VERIFY_EJECT = 20
AYA_MIN_EJECT = 3

_reset = True
_cfg = {
    "mode": None,
    "red": 0,
    "green": 0,
    "blue": 0,
    "brightness": 1,
}

CONTROLLER_POWER = (
    "/sys/class/firmware-attributes/ayaneo-ec/attributes/controller_power/current_value",
    "/sys/devices/platform/ayaneo-ec/controller_power",
)
CONTROLLER_MODULES = [
    "/sys/class/firmware-attributes/ayaneo-ec/attributes/controller_modules/current_value",
    "/sys/devices/platform/ayaneo-ec/controller_modules",
]


def write_cmd(dev, r: bytes, timeout: int = AYA_TIMEOUT):
    if DEBUG_MODE:
        logger.info(f"Send: {r[1:].hex()}")
    dev.write(r)

    # Read response
    for _ in range(AYA_ATTEMPTS):
        res = dev.read(timeout=timeout)
        if DEBUG_MODE:
            logger.info(f"Recv: {res.hex() if res else 'None'}")
        # Match IDs
        if res and res[3] == r[4]:
            break

    return res


LEFT_MODULES = {
    0x02: _("Cross Film / Joystick"),
    0x04: _("Cross / Joystick"),
    0x06: _("Cross / Touchpad"),
    0x08: _("Direction / Joystick"),
    0x42: _("Joystick / Cross Film"),
    0x44: _("Joystick / Cross"),
    0x46: _("Touchpad / Cross"),
    0x48: _("Joystick / Direction"),
}

RIGHT_MODULES = {
    0x10: _("ABXY \\ Joystick"),
    0x12: _("ABXY \\ Touchpad"),
    0x14: _("ABXYCZ"),
    0x16: _("ABXY Film \\ Joystick"),
    0x50: _("Joystick \\ ABXY"),
    0x52: _("Touchpad \\ ABXY"),
    0x54: _("ABXYCZ [R]"),
    0x56: _("Joystick \\ ABXY Film"),
}

MODULE_UKN = _("Unknown")
MODULE_EJECTING = _("Ejecting...")
MODULE_ACTIVATING = _("Activating...")
MODULE_UNPOWERED = _("Unpowered")
MODULE_DISCONNECTED = _("Disconnected")
MODULE_DISABLED = _("Paused")


def write_power(enable: bool):
    for fn in CONTROLLER_POWER:
        if os.path.exists(fn):
            with open(fn, "w") as f:
                f.write(
                    "on"
                    if enable and "firmware-attributes" in fn
                    else (
                        "1" if enable else "off" if "firmware-attributes" in fn else "0"
                    )
                )
            logger.info(f"Controller power turned {'on' if enable else 'off'}.")
            return True
    return False


def read_power():
    for fn in CONTROLLER_POWER:
        if os.path.exists(fn):
            with open(fn, "r") as f:
                power = f.read().strip()
            return power in ("on", "1")
    return None


def read_modules():
    for fn in CONTROLLER_MODULES:
        if os.path.exists(fn):
            with open(fn, "r") as f:
                modules = f.read().strip()
            return modules
    return None


class Ayaneo3Hidraw(GenericGamepadHidraw):

    def __init__(self, *args, vibration: str, config: Config, emit, **kwargs):
        super().__init__(*args, **kwargs)
        self.vibration = vibration
        self.config = config
        self.emit = emit

    def init(self):
        if not self.dev:
            return

        for cmd in AYA3_INIT:
            write_cmd(self.dev, to_bytes(cmd))

    def cfg(self):
        # Send init immediately after sleep
        if _cfg["mode"] is not None:
            self.send_cfg(reset=False)

    def send_cfg(self, reset: bool = False, left: bool = False, right: bool = False):
        cmds = get_cfg_commands(
            rgb_mode=_cfg["mode"] or "disabled",
            r=_cfg["red"],
            g=_cfg["green"],
            b=_cfg["blue"],
            brightness=_cfg["brightness"],
            reset=reset,
            eject_left=left,
            eject_right=right,
            vibration=self.vibration,  # type: ignore[call-arg]
        )
        for cmd in cmds:
            write_cmd(self.dev, cmd)

    def save(self):
        if not self.dev:
            return

        # Save current config
        write_cmd(self.dev, AYA_SAVE, timeout=AYA_SAVE_TIMEOUT)

    def reset(self):
        self.send_cfg(reset=True)
        time.sleep(0.5)
        self.send_cfg(reset=False)

    def check(self, init=False):
        for i in range(AYA_ATTEMPTS):
            res = write_cmd(self.dev, AYA_CHECK)
            if not res:
                time.sleep(0.3)
                continue
            left = res[32]
            right = res[33]
            if not right or not left:
                time.sleep(0.3)
                continue

            # left_module = left & 0x3F
            # left_rotated = left >> 6
            vl = LEFT_MODULES.get(left, MODULE_UKN)
            refresh = self.config.get("info_left", None) != vl
            self.config["info_left"] = vl
            # right_module = right & 0x3F
            # right_rotated = right >> 6
            vr = RIGHT_MODULES.get(right, MODULE_UKN)
            refresh |= self.config.get("info_right", None) != vr
            self.config["info_right"] = vr
            if refresh:
                self.emit([])
            break

        if res and res[18] == 1:
            logger.info("Switching into custom mode.")
            time.sleep(1)
            write_cmd(self.dev, AYA_CUSTOM)
            time.sleep(2)
            write_cmd(self.dev, AYA_CHECK)
            return True
        else:
            logger.info("Controller module check passed.")

    def eject(self, left: bool = False, right: bool = False):
        if left:
            self.config["info_left"] = MODULE_EJECTING
        if right:
            self.config["info_right"] = MODULE_EJECTING
        if left or right:
            self.emit([])

        self.send_cfg(reset=False, left=left, right=right)
        self.handle_eject()

    def handle_eject(self, throw=False):
        start = time.perf_counter()
        for _ in range(AYA_VERIFY_EJECT):
            res = write_cmd(self.dev, AYA_CHECK)
            time.sleep(0.4)
            # wait after last check too
            if res and res[19] & ~0x11 == 0:
                logger.info("Eject verified.")
                break

        time_waited = time.perf_counter() - start
        if time_waited < AYA_MIN_EJECT:
            time.sleep(AYA_MIN_EJECT - time_waited)

        if write_power(False):
            time.sleep(0.5)
        else:
            logger.warning("Kernel driver for modules is missing. Sleeping.")
            os.system("systemctl suspend")

        if throw:
            raise RuntimeError("Turned off controller. Restarting.")

    def consume(self, events):
        if not self.dev:
            return

        got_rgb = False
        for ev in events:
            if ev["type"] != "led":
                continue

            mode = ev["mode"]

            if mode not in ["disabled", "solid", "pulse", "rainbow"]:
                logger.error(f"Invalid RGB mode: {mode}")
                continue

            # Only send RGB command if colors change
            got_rgb = (
                _cfg["mode"] != mode
                or _cfg["red"] != ev["red"]
                or _cfg["green"] != ev["green"]
                or _cfg["blue"] != ev["blue"]
                or _cfg["brightness"] != ev.get("brightness", 1)
            )
            _cfg["mode"] = mode
            _cfg["red"] = ev["red"]
            _cfg["green"] = ev["green"]
            _cfg["blue"] = ev["blue"]
            _cfg["brightness"] = ev.get("brightness", 1)

        if got_rgb:
            self.send_cfg()


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
    others: Config,
):
    first = True
    first_disabled = True
    init = time.perf_counter()
    repeated_fail = False

    if others.get("info_left", None) is None:
        others["info_left"] = MODULE_UNPOWERED
    if others.get("info_right", None) is None:
        others["info_right"] = MODULE_UNPOWERED

    while not should_exit.is_set():
        if conf["controller_mode.mode"].to(str) == "disabled":
            # Disable module handling
            others["info_left"] = MODULE_DISABLED
            others["info_right"] = MODULE_DISABLED
            others.pop("reset", False)
            others.pop("pop", None)

            time.sleep(ERROR_DELAY)
            if first_disabled:
                UInputDevice.close_volume_cached()
                unhide_all()
            first_disabled = False
            continue
        else:
            first_disabled = True

        found_device = True
        modules = read_modules()
        if modules is not None:
            # Check if status should be refreshed
            refresh = False
            if modules == "both":
                found_device = True

                # Update status
                if others.get("info_right", None) in (
                    MODULE_DISCONNECTED,
                    MODULE_UNPOWERED,
                ):
                    others["info_right"] = MODULE_ACTIVATING
                    refresh = True
                if others.get("info_left", None) in (
                    MODULE_DISCONNECTED,
                    MODULE_UNPOWERED,
                ):
                    others["info_left"] = MODULE_ACTIVATING
                    refresh = True
            else:
                if first:
                    logger.info(
                        f"Controller modules are set to '{modules}', waiting for 'both' to be connected."
                    )
                found_device = False

                # Update status
                if modules != "left":
                    refresh |= others.get("info_left", None) != MODULE_DISCONNECTED
                    others["info_left"] = MODULE_DISCONNECTED
                if modules != "right":
                    refresh |= others.get("info_right", None) != MODULE_DISCONNECTED
                    others["info_right"] = MODULE_DISCONNECTED
                if modules == "right":
                    if others.get("info_right", None) == MODULE_DISCONNECTED:
                        others["info_right"] = MODULE_UNPOWERED
                        refresh = True
                if modules == "left":
                    if others.get("info_left", None) == MODULE_DISCONNECTED:
                        others["info_left"] = MODULE_UNPOWERED
                        refresh = True

                # Turning off controller power if not connected
                write_power(False)

            if refresh:
                emit([])

        pop = others.pop("pop", None)
        if pop:
            others.pop("reset", False)
        if found_device or pop:
            power = read_power()
            if power is False:
                logger.info(
                    f"Controller power is set to '{power}', powering on controller."
                )
                write_power(True)
                time.sleep(1)

        if pop:
            logger.info(f"Popping controller module {pop}.")
            d_vend = Ayaneo3Hidraw(
                vid=[AYA_VID],
                pid=[AYA_PID],
                required=True,
                application=[0xFF000001],
                vibration=conf.get("vibration", "medium"),
                config=others,
                emit=emit,
            )
            try:
                d_vend.open()
                d_vend.eject(
                    left=pop in ["both", "left"], right=pop in ["both", "right"]
                )
                found_device = False
            except Exception as e:
                logger.error(f"Failed to pop controller module {pop} with error:\n{e}")
                if DEBUG_MODE:
                    raise e
            finally:
                d_vend.close(True)

        try:
            if not bool(enumerate_evs(vid=GAMEPAD_VID, pid=GAMEPAD_PID)):
                found_device = False
        except Exception:
            logger.warning("Failed finding device, skipping check.")
            time.sleep(LONGER_ERROR_DELAY)

        if not found_device:
            if first:
                logger.info("Controller not found. Waiting...")
            time.sleep(FIND_DELAY)
            first = False
            continue

        try:
            logger.info("Launching emulated controller.")
            updated.clear()
            init = time.perf_counter()
            reset = others.pop("reset", False)
            controller_loop(
                conf.copy(), should_exit, updated, dconf, emit, others, reset
            )
            repeated_fail = False
        except Exception as e:
            failed_fast = init + LONGER_ERROR_MARGIN > time.perf_counter()
            sleep_time = (
                LONGER_ERROR_DELAY if repeated_fail and failed_fast else ERROR_DELAY
            )
            repeated_fail = failed_fast
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {sleep_time}s."
            )
            first = True
            # Raise exception
            if DEBUG_MODE:
                raise e
            time.sleep(sleep_time)

    # Unhide all devices before exiting and close keyboard cache
    UInputDevice.close_volume_cached()
    unhide_all()


def controller_loop(
    conf: Config,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
    emit: Emitter,
    others: Config,
    reset: bool = False,
):
    global _reset

    debug = DEBUG_MODE
    dtype = dconf.get("type", "generic")
    dgyro = dconf.get("display_gyro", True)

    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        None,
        dgyro and conf["imu"].to(bool),
        emit=emit,
        rgb_modes=(
            {"disabled": [], "solid": ["color"], "pulse": ["color"], "rainbow": []}
            if dconf.get("rgb", False)
            else None
        ),
        rgb_init_times=1,
        extra_buttons=dconf.get("extra_buttons", "dual"),
    )
    motion = d_params.get("uses_motion", True)

    # Imu
    if dgyro:
        d_imu = CombinedImu(
            conf["imu_hz"].to(int),
            get_gyro_state(conf["imu_axis"], dconf.get("mapping", DEFAULT_MAPPINGS)),
        )
    else:
        d_imu = None
    d_timer = HrtimerTrigger(conf["imu_hz"].to(int), [HrtimerTrigger.IMU_NAMES])

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[GAMEPAD_VID],
        pid=[GAMEPAD_PID],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
        btn_map={**XBOX_BUTTON_MAP, EC("BTN_MODE"): "share"},
    )

    d_kbd_1 = GenericGamepadEvdev(
        vid=[KBD_VID],
        pid=[KBD_PID],
        required=False,
        grab=False,
        btn_map={
            EC("KEY_F23"): "mode",
        },
    )
    d_kbd_2 = GenericGamepadEvdev(
        vid=[AYA_VID],
        pid=[AYA_PID],
        required=True,
        grab=True,
        capabilities={EC("EV_KEY"): [EC("KEY_F21")]},
        btn_map={
            EC("KEY_F24"): "keyboard",
            EC("KEY_D"): "keyboard",
            EC("KEY_F21"): "extra_l2",
            EC("KEY_F22"): "extra_r2",
            EC("KEY_L"): "extra_l1",
            EC("KEY_R"): "extra_r1",
        },
    )
    d_vend = Ayaneo3Hidraw(
        vid=[AYA_VID],
        pid=[AYA_PID],
        required=True,
        application=[0xFF000001],
        vibration=conf.get("vibration", "medium"),
        config=others,
        emit=emit,
    )

    match conf.get("swap_guide", "oem"):
        case "traditional":
            swap_guide = "aya_traditional"
        case "traditional_rev":
            swap_guide = "aya_traditional_rev"
        case _:
            swap_guide = None

    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        share_to_qam=True,
        select_reboots=conf.get("select_reboots", False),
        nintendo_mode=conf["nintendo_mode"].to(bool),
        emit=emit,
        params=d_params,
        startselect_chord=conf.get("main_chords", "disabled"),
        keyboard_is="qam",
        qam_hhd="full",
        swap_guide=swap_guide,
    )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    if motion:
        REPORT_FREQ_MAX = max(REPORT_FREQ_MAX, conf["imu_hz"].to(float))

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        prepare(d_xinput)
        if motion and d_imu:
            start_imu = True
            if dconf.get("hrtimer", False):
                start_imu = d_timer.open()
            if start_imu:
                prepare(d_imu)
        prepare(d_kbd_1)
        if d_kbd_2:
            prepare(d_kbd_2)
        for d in d_producers:
            prepare(d)

        prepare(d_vend)
        changed_mode = d_vend.check()
        if reset or _reset or changed_mode:
            d_vend.init()
            if not changed_mode:
                d_vend.reset()
            d_vend.save()
            reset = False
            _reset = False
        else:
            d_vend.cfg()

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                d_xinput.consume(evs)

            if d_vend:
                d_vend.consume(evs)
            for d in d_outs:
                d.consume(evs)

            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        # d_vend.close(not updated.is_set())
        try:
            d_timer.close()
        except Exception as e:
            logger.error(f"Error while closing device '{d}' with exception:\n{e}")
            if debug:
                raise e
        for d in reversed(devs):
            try:
                d.close(not updated.is_set())
            except Exception as e:
                logger.error(f"Error while closing device '{d}' with exception:\n{e}")
                if debug:
                    raise e



from threading import Event as TEvent, Thread
from typing import Any, Sequence

from hhd.plugins import (
    Config,
    Context,
    Emitter,
    Event,
    HHDPlugin,
    load_relative_yaml,
    get_outputs_config,
    get_limits_config,
    fix_limits,
)
from hhd.plugins.settings import HHDSettings


class RogAllyControllersPlugin(HHDPlugin):
    name = "rog_ally_controllers"
    priority = 18
    log = "ally"

    def __init__(self, ally_x: bool = False, xbox: bool = False) -> None:
        self.t = None
        self.should_exit = None
        self.updated = TEvent()
        self.woke_up = None
        self.started = False
        self.t = None
        self.ally_x = ally_x
        self.xbox = xbox

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        self.emit = emit
        self.context = context
        self.prev = None

    def settings(self) -> HHDSettings:
        from .base import LIMIT_DEFAULTS

        base = {"controllers": {"rog_ally": load_relative_yaml("controllers.yml")}}
        base["controllers"]["rog_ally"]["children"]["controller_mode"].update(
            get_outputs_config(
                can_disable=False, extra_buttons="dual", noob_default=not self.xbox
            )
        )
        base["controllers"]["rog_ally"]["children"]["limits"] = get_limits_config(
            LIMIT_DEFAULTS(self.ally_x)
        )
        
        if not self.xbox:
            del base["controllers"]["rog_ally"]["children"]["swap_xbox"]

        return base

    def update(self, conf: Config):
        from .base import LIMIT_DEFAULTS

        fix_limits(conf, "controllers.rog_ally.limits", LIMIT_DEFAULTS(self.ally_x))

        new_conf = conf["controllers.rog_ally"]
        if new_conf == self.prev:
            return
        if self.prev is None:
            self.prev = new_conf
        else:
            self.prev.update(new_conf.conf)

        self.updated.set()
        self.start(self.prev)

    def start(self, conf):
        from .base import plugin_run

        if self.started:
            return
        self.started = True

        self.close()
        self.should_exit = TEvent()
        self.woke_up = TEvent()
        self.t = Thread(
            target=plugin_run,
            args=(
                conf,
                self.emit,
                self.context,
                self.should_exit,
                self.updated,
                self.ally_x,
                self.xbox,
                self.woke_up,
            ),
        )
        self.t.start()

    def close(self):
        if not self.should_exit or not self.t:
            return
        self.should_exit.set()
        self.t.join()
        self.should_exit = None
        self.t = None

    def notify(self, events: Sequence[Event]):
        for ev in events:
            if ev["type"] == "special" and ev.get("event", None) == "wakeup":
                if self.woke_up:
                    self.woke_up.set()


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    # Match just product number, should be enough for now
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        # Different variants of the ally can have an additional _RC71L or not
        dmi = f.read().strip()

    # First gen ally
    # ROG Ally RC71L_Action or something else
    if "ROG Ally RC71L" in dmi:
        return [RogAllyControllersPlugin()]

    # Ally X
    # ROG Ally X RC72LA_RC72LA_000123206
    if "ROG Ally X RC72LA" in dmi:
        return [RogAllyControllersPlugin(ally_x=True)]

    # Xbox Ally
    # ROG Xbox Ally RC73YA_RC73YA
    # Xbox Ally X
    # ROG Xbox Ally X RC73XA_RC73XA
    if "ROG Xbox Ally X RC73X" in dmi or "ROG Xbox Ally RC73Y" in dmi:
        return [RogAllyControllersPlugin(ally_x=True, xbox=True)]

    return []



from threading import Event, Thread
from typing import Sequence

from hhd.plugins import (
    Config,
    Context,
    Emitter,
    HHDPlugin,
    get_gyro_config,
    get_outputs_config,
    load_relative_yaml,
)
from hhd.plugins.settings import HHDSettings

from .const import CONFS, DEFAULT_MAPPINGS


class AyaneoControllersPlugin(HHDPlugin):
    name = "ayaneo_controllers"
    priority = 18
    log = "ayac"

    def __init__(self, dmi: str, dconf: dict) -> None:
        self.t = None
        self.should_exit = None
        self.updated = Event()
        self.started = False
        self.t = None

        self.dmi = dmi
        self.dconf = dconf
        self.magic_modules = dconf.get("magic_modules", False)
        self.name = f"ayaneo_controllers@'{dconf.get('name', 'ukn')}'"

        self.config = Config()

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        self.emit = emit
        self.context = context
        self.prev = None

    def settings(self) -> HHDSettings:
        if self.magic_modules:
            base = {
                "controllers": {
                    "ayaneo": load_relative_yaml("controllers.yml"),
                },
                "magic_modules": {
                    "magic_modules": load_relative_yaml("modules.yml"),
                },
            }
        else:
            base = {
                "controllers": {
                    "ayaneo": load_relative_yaml("controllers.yml"),
                }
            }

        base["controllers"]["ayaneo"]["children"]["controller_mode"].update(
            get_outputs_config(
                can_disable=True,
                has_leds=self.dconf.get("rgb", False),
                start_disabled=self.dconf.get("untested", False),
                extra_buttons=self.dconf.get("extra_buttons", "quad"),
                noob_default=False,
            )
        )

        if self.dconf.get("display_gyro", True):
            base["controllers"]["ayaneo"]["children"]["imu_axis"] = get_gyro_config(
                self.dconf.get("mapping", DEFAULT_MAPPINGS)
            )
        else:
            del base["controllers"]["ayaneo"]["children"]["imu_axis"]
            del base["controllers"]["ayaneo"]["children"]["imu"]

        return base

    def update(self, conf: Config):
        new_conf = conf["controllers.ayaneo"]

        if self.magic_modules:
            pop_both = conf.get_action("magic_modules.magic_modules.pop_both")
            pop_left = conf.get_action("magic_modules.magic_modules.pop_left")
            pop_right = conf.get_action("magic_modules.magic_modules.pop_right")
            reset = conf.get_action("magic_modules.magic_modules.reset")

            conf["magic_modules.magic_modules.info_right"] = self.config.get(
                "info_right", None
            )
            conf["magic_modules.magic_modules.info_left"] = self.config.get(
                "info_left", None
            )
        else:
            pop_both = False
            pop_left = False
            pop_right = False
            reset = False

        if (
            new_conf == self.prev
            and not pop_both
            and not pop_left
            and not pop_right
            and not reset
        ):
            return
        if self.prev is None:
            self.prev = new_conf
        else:
            self.prev.update(new_conf.conf)

        if pop_both:
            pop = "both"
        elif pop_left:
            pop = "left"
        elif pop_right:
            pop = "right"
        else:
            pop = None

        if pop_both or pop_left or pop_right or reset:
            self.started = False
        else:
            self.updated.set()
        self.start(self.prev, pop=pop, reset=reset)

    def start(self, conf, pop=None, reset=False):
        from .base import plugin_run

        if self.started:
            return
        self.started = True

        self.close()
        self.should_exit = Event()
        self.config["pop"] = pop
        self.config["reset"] = reset
        self.t = Thread(
            target=plugin_run,
            args=(
                conf,
                self.emit,
                self.context,
                self.should_exit,
                self.updated,
                self.dconf,
                self.config,
            ),
        )
        self.t.start()

    def close(self):
        if not self.should_exit or not self.t:
            return
        self.should_exit.set()
        self.t.join()
        self.should_exit = None
        self.t = None


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    # Match just product name
    # if a device exists here its officially supported
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        dmi = f.read().strip()

    dconf = CONFS.get(dmi, None)
    if dconf:
        return [AyaneoControllersPlugin(dmi, dconf)]

    return []



import logging
import select
from threading import Event as TEvent
from typing import Any, Sequence

from hhd.plugins import Emitter, Event
from pyroute2 import AcpiEventSocket  # type: ignore

logger = logging.getLogger(__name__)

EVENT_MATCHES: Sequence[tuple[dict[str, Any], str]] = [
    ({"device_class": b"ac_adapter", "data": 0}, "dc"),
    ({"device_class": b"ac_adapter", "data": 256}, "ac"),
    ({"device_class": b"battery"}, "battery"),
    ({"device_class": b"button/power"}, "powerbutton"),
    # Legion GO TDP event
    ({"bus_id": b"D320289E-8FEA-"}, "tdp"),
    ({"device_class": b"wmi", "bus_id": b"PNP0C14:01"}, "tdp"),  # Legion go
    ({"device_class": b"wmi", "bus_id": b"PNP0C14:00"}, "tdp"),  # Legion go s
    # GPD Force hibernate thermal event
    # , 'type': 0xf100, 'data': 0x0100 ignore these attrs for now...
    # ({"device_class": b"thermal_zone", "bus_id": b"LNXTHERM:00"}, "hibernate-thermal"),
]

GUARD_DELAY = 0.5


def loop_process_events(emit: Emitter, should_exit: TEvent):
    acpi = AcpiEventSocket()
    logger.info(f"Starting ACPI Event handler.")

    while not should_exit.is_set():
        # FIXME: Uses unofficial API, find the correct way.
        try:
            sock = acpi.asyncore.socket
        except:
            sock = acpi._sock
        r, _, _ = select.select([sock.fileno()], [], [], GUARD_DELAY)

        if r:
            for message in acpi.get():
                ev = message.get("ACPI_GENL_ATTR_EVENT", None)

                if not ev:
                    continue

                found = False
                for match, etype in EVENT_MATCHES:
                    matches = True
                    for k, v in match.items():
                        if k not in ev or ev[k] != v:
                            matches = False
                            break

                    if matches:
                        if etype not in ("battery", "powerbutton"):
                            emit({"type": "acpi", "event": etype})  # type: ignore
                        found = True
                        break

                if not found:
                    logger.info(f"ACPI event: {ev}")
